class NintendoException(Exception):
    """Base class for all Nintendo exceptions."""

    pass


class FTokenException(Exception):
    """Base class for all Imink exceptions."""

    pass


class SplatnetException(Exception):
    """Base class for all Splatnet exceptions."""

    pass
